--------------------------------------------------------------------
--	Fermat theorem algorithm
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;

entity fermat_div is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  clk, reset, start: in std_logic;
  done: out std_logic;
  z: out std_logic_vector(k-1 downto 0)
);
end fermat_div;

architecture circuit of fermat_div is
  component montgomery_prod is
  port (
    x, y: in std_logic_vector(k-1 downto 0);
    clk, reset, start: in std_logic;
    done: out std_logic;
    z: out std_logic_vector(k-1 downto 0)
  );
  end component;
  signal first_operand, second_operand, mp, Tx, Ty, e: std_logic_vector(k-1 downto 0);
  signal sel: std_logic_vector(2 downto 0);
  signal first_step, ce_Tx, ce_Ty, ce_e, start_mp, done_mp, load, shift: std_logic;
  signal count: std_logic_vector(logk-1 downto 0);
  subtype states is natural range 0 to 20;
  signal current_state: states;
begin
  with sel select 
    first_operand <= y when "000", 
	                   x when "001", 
	                   e when others;
  with sel select 
    second_operand <= exp_2k when "000"|"001", 
	                         e when "010", 
								          Ty when "011", 
                          Tx when "100", 
								          conv_std_logic_vector(1, k) when others;
  
  montg_prod: montgomery_prod port map(first_operand, second_operand, clk, reset, start_mp, done_mp, mp);

  register_e: process(clk)
  begin
  if clk'event and clk = '1' then
    if first_step = '1' then 
	    e <= exp_k;			 
    elsif ce_e = '1' then 
      e <= mp; 
    end if;
  end if;
  end process register_e;

  register_Tx: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_Tx = '1' then 
	   Tx <= mp; 
    end if;
  end if;
  end process register_Tx;

  register_Ty: process(clk)
  begin
  if clk'event and clk = '1' then
    if ce_Ty = '1' then 
	   Ty <= mp; 
	 end if;
  end if;
  end process register_Ty;

  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
	   count <= conv_std_logic_vector(k-1, logk);
    elsif shift = '1' then 
	   count <= count-1;
    end if;
  end if;
  end process counter;

  control_unit: process(clk, reset, current_state, count)
  begin
  case current_state is
    when 0 to 1 => sel <= "000"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '1';

    when 2 => sel <= "000"; start_mp <= '1'; first_step <= '1'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 3 => sel <= "000"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 4 => sel <= "000"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '1'; load <= '0'; shift <= '0'; done <= '0';

    when 5 => sel <= "001"; start_mp <= '1'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 6 => sel <= "001"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 7 => sel <= "001"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '1'; ce_Ty <= '0'; load <= '1'; shift <= '0'; done <= '0';

    when 8 => sel <= "010"; start_mp <= '1'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 9 => sel <= "010"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 10 => sel <= "010"; start_mp <= '0'; first_step <= '0'; ce_e <= '1'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';

    when 11 => sel <= "011"; start_mp <= '1'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 12 => sel <= "011"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 13 => sel <= "011"; start_mp <= '0'; first_step <= '0'; ce_e <= '1'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';

    when 14 => sel <= "010"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '1'; done <= '0';

    when 15 => sel <= "100"; start_mp <= '1'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 16 => sel <= "100"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 17 => sel <= "100"; start_mp <= '0'; first_step <= '0'; ce_e <= '1'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';

    when 18 => sel <= "101"; start_mp <= '1'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 19 => sel <= "101"; start_mp <= '0'; first_step <= '0'; ce_e <= '0'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';
    when 20 => sel <= "101"; start_mp <= '0'; first_step <= '0'; ce_e <= '1'; ce_Tx <= '0'; ce_Ty <= '0'; load <= '0'; shift <= '0'; done <= '0';

  end case;
  if reset = '1' then 
    current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= current_state + 1; end if;
      when 1 => if start = '1' then current_state <= current_state + 1; end if;
      when 2 => current_state <= current_state + 1; --Ty := MP(Y, Exp_2K)
      when 3 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 4 => current_state <= current_state + 1;
      when 5 => current_state <= current_state + 1; --Tx := MP(X, Exp_2K)
      when 6 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 7 => current_state <= current_state + 1;
      when 8 => current_state <= current_state + 1; --E := MP(E, E)
      when 9 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 10 => if p_minus2(conv_integer(count)) = '1' then 
                   current_state <= current_state + 1; 
                 else current_state <= 14; 
                 end if;
      when 11 => current_state <= current_state + 1; --E := MP(E, Ty)
      when 12 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 13 => current_state <= current_state + 1; 

      when 14 => if count = 0 then current_state <= current_state + 1; else current_state <= 8; end if;

      when 15 => current_state <= current_state + 1;
      when 16 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 17 => current_state <= current_state + 1;
      when 18 => current_state <= current_state + 1;
      when 19 => if done_mp = '1' then current_state <= current_state + 1; end if;
      when 20 => current_state <= 0;
    end case;
  end if;
  end process;
  
  z <= e;
end circuit;






