--------------------------------------------------------------------
--	Fermat theorem algorithm. Montgomery Product Data Path
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;

entity montg_data_path is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  clk, load, shift, ce_product, clear_product: in std_logic;
  z: out std_logic_vector(k-1 downto 0)
);
end montg_data_path;

architecture rtl of montg_data_path is
signal reg_x: std_logic_vector(k-1 downto 0);
signal a, a_plus_p, dif: std_logic_vector(k+1 downto 0);
signal product, next_product: std_logic_vector(k downto 0);
signal x_i: std_logic;
begin
  with x_i select a <= ('0'&product) + ("00"&y) when '1', ('0'&product) when others;
  with a(0) select a_plus_p <= a + ("00"&p) when '1', a when others;

  next_product(k downto 0) <= a_plus_p(k+1 downto 1);

  shift_register: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
	    reg_x <= x;
    elsif shift = '1' then
      reg_x(k-1 downto 0) <= '0' & reg_x(k-1 downto 1); 
    end if;
  end if;
  end process shift_register;
  
  x_i <= reg_x(0);

  register_product: process(clk)
  begin
  if clk'event and clk = '1' then
    if clear_product = '1' then 
	    product <= (others => '0');
    elsif ce_product = '1' then 
	    product <= next_product;
    end if;
  end if;
  end process register_product;
  
  dif <= ('0' & product) - ("00" & p);

  with dif(k+1) select z <= product(k-1 downto 0) when '1', dif(k-1 downto 0) when others;

end rtl;

