--------------------------------------------------------------------
--	Fermat theorem algorithm. Montgomery Product (use Montg_Data_Path)
--------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.my_package.all;

entity montgomery_prod is
port (
  x, y: in std_logic_vector(k-1 downto 0);
  clk, reset, start: in std_logic;
  done: out std_logic;
  z: out std_logic_vector(k-1 downto 0)
);
end montgomery_prod;

architecture circuit of montgomery_prod is
  component montg_data_path is
  port (
    x, y: in std_logic_vector(k-1 downto 0);
    clk, load, shift, ce_product, clear_product: in std_logic;
    z: out std_logic_vector(k-1 downto 0)
  );
  end component;
  signal load, shift, ce_product, clear_product: std_logic;
  signal count: std_logic_vector(logk-1 downto 0);
  subtype states is natural range 0 to 3;
  signal current_state: states;
begin
  
  data_path: montg_data_path port map(x, y, clk, load, shift, ce_product, clear_product, z);
  
  counter: process(clk)
  begin
  if clk'event and clk = '1' then
    if load = '1' then 
	    count <= conv_std_logic_vector(k-1, logk);
    elsif shift = '1' then 
	    count <= count-1;
    end if;
  end if;
  end process counter;

  control_unit: process(clk, reset, current_state, count)
  begin
  case current_state is
    when 0 to 1 => load <= '0'; shift <= '0'; ce_product <= '0'; clear_product <= '0'; done <= '1';
    when 2 => load <= '1'; shift <= '0'; ce_product <= '0'; clear_product <= '1'; done <= '0';
    when 3 => load <= '0'; shift <= '1'; ce_product <= '1'; clear_product <= '0'; done <= '0';
  end case;

  if reset = '1' then current_state <= 0;
  elsif clk'event and clk = '1' then
    case current_state is
      when 0 => if start = '0' then current_state <= current_state + 1; end if;
      when 1 => if start = '1' then current_state <= current_state + 1; end if;
      when 2 => current_state <= current_state + 1;
      when 3 => if count = 0 then current_state <= 0; end if;
    end case;
   end if;
  end process;
end circuit;
